@@ -3,17 +3,19 @@
 
 from abc import ABC, abstractmethod
 from copy import deepcopy
+from functools import partial
 from math import ceil
 from typing import Any, Callable, Dict, List, Optional, Tuple, Union
 from warnings import warn
 
 import numpy as np
 import torch
+import torch.distributions.transforms as torch_tf
 from pyro.infer.mcmc import HMC, NUTS
 from pyro.infer.mcmc.api import MCMC
-from torch import Tensor, float32
+from torch import Tensor
 from torch import multiprocessing as mp
-from torch import nn, optim
+from torch import nn
 
 from sbi import utils as utils
 from sbi.mcmc import (
@@ -25,9 +27,13 @@ from sbi.mcmc import (
     sir,
 )
 from sbi.types import Array, Shape
-from sbi.utils.sbiutils import check_dist_class, check_warn_and_setstate
+from sbi.utils.sbiutils import (
+    check_warn_and_setstate,
+    mcmc_transform,
+    optimize_potential_fn,
+    rejection_sample,
+)
 from sbi.utils.torchutils import (
-    BoxUniform,
     ScalarFloat,
     atleast_2d_float32_tensor,
     ensure_theta_batched,
@@ -50,8 +56,10 @@ class NeuralPosterior(ABC):
         neural_net: nn.Module,
         prior,
         x_shape: torch.Size,
+        sample_with: str,
         mcmc_method: str = "slice_np",
         mcmc_parameters: Optional[Dict[str, Any]] = None,
+        rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
         device: str = "cpu",
     ):
         """
@@ -60,6 +68,8 @@ class NeuralPosterior(ABC):
             neural_net: A classifier for SNRE, a density estimator for SNPE and SNL.
             prior: Prior distribution with `.log_prob()` and `.sample()`.
             x_shape: Shape of the simulator data.
+            sample_with: Method to use for sampling from the posterior. Must be one of
+                [`mcmc` | `rejection`].
             mcmc_method: Method used for MCMC sampling, one of `slice_np`, `slice`,
                 `hmc`, `nuts`. Currently defaults to `slice_np` for a custom numpy
                 implementation of slice sampling; select `hmc`, `nuts` or `slice` for
@@ -72,6 +82,15 @@ class NeuralPosterior(ABC):
                 will draw init locations from prior, whereas `sir` will use Sequential-
                 Importance-Resampling. Init strategies may have their own keywords
                 which can also be set from `mcmc_parameters`.
+            rejection_sampling_parameters: Dictionary overriding the default parameters
+                for rejection sampling. The following parameters are supported:
+                `proposal` as the proposal distribtution.
+                `max_sampling_batch_size` as the batchsize of samples being drawn from
+                the proposal at every iteration. `num_samples_to_find_max` as the
+                number of samples that are used to find the maximum of the
+                `potential_fn / proposal` ratio. `num_iter_to_find_max` as the number
+                of gradient ascent iterations to find the maximum of that ratio. `m` as
+                multiplier to that ratio.
             device: Training device, e.g., cpu or cuda.
         """
         if method_family in ("snpe", "snle", "snre_a", "snre_b"):
@@ -83,6 +102,8 @@ class NeuralPosterior(ABC):
 
         self.set_mcmc_method(mcmc_method)
         self.set_mcmc_parameters(mcmc_parameters)
+        self.set_sample_with(sample_with)
+        self.set_rejection_sampling_parameters(rejection_sampling_parameters)
 
         self._leakage_density_correction_factor = None  # Correction factor for SNPE.
         self._mcmc_init_params = None
@@ -129,11 +150,45 @@ class NeuralPosterior(ABC):
         Returns:
             `NeuralPosterior` that will use a default `x` when not explicitly passed.
         """
-        self._x = process_x(x, self._x_shape, allow_iid_x=self._allow_iid_x)
+        self._x = process_x(x, self._x_shape, allow_iid_x=self._allow_iid_x).to(
+            self._device
+        )
         self._num_iid_trials = self._x.shape[0]
 
         return self
 
+    @property
+    def sample_with(self) -> str:
+        """
+        Return `True` if NeuralPosterior instance should use MCMC in `.sample()`.
+        """
+        return self._sample_with
+
+    @sample_with.setter
+    def sample_with(self, value: str) -> None:
+        """See `set_sample_with`."""
+        self.set_sample_with(value)
+
+    def set_sample_with(self, sample_with: str) -> "NeuralPosterior":
+        """Set the sampling method for the `NeuralPosterior`.
+
+        Args:
+            sample_with: The method to sample with.
+
+        Returns:
+            `NeuralPosterior` for chainable calls.
+
+        Raises:
+            ValueError: on attempt to turn off MCMC sampling for family of methods that
+                do not support rejection sampling.
+        """
+        if sample_with not in ("mcmc", "rejection"):
+            raise NameError(
+                "The only implemented sampling methods are `mcmc` and `rejection`."
+            )
+        self._sample_with = sample_with
+        return self
+
     @property
     def mcmc_method(self) -> str:
         """Returns MCMC method."""
@@ -188,6 +243,39 @@ class NeuralPosterior(ABC):
         self._mcmc_parameters = parameters
         return self
 
+    @property
+    def rejection_sampling_parameters(self) -> dict:
+        """Returns rejection sampling parameters."""
+        if self._rejection_sampling_parameters is None:
+            return {}
+        else:
+            return self._rejection_sampling_parameters
+
+    @rejection_sampling_parameters.setter
+    def rejection_sampling_parameters(self, parameters: Dict[str, Any]) -> None:
+        """See `set_rejection_sampling_parameters`."""
+        self.set_rejection_sampling_parameters(parameters)
+
+    def set_rejection_sampling_parameters(
+        self, parameters: Dict[str, Any]
+    ) -> "NeuralPosterior":
+        """Sets parameters for rejection sampling and returns `NeuralPosterior`.
+
+        Args:
+            parameters: Dictonary overriding the default parameters
+                for rejection sampling. The following parameters are supported:
+                `proposal` as the proposal distribtution. `num_samples_to_find_max`
+                as the number of samples that are used to find the maximum of the
+                `potential_fn / proposal` ratio. `m` as multiplier to that ratio.
+                `sampling_batch_size` as the batchsize of samples being drawn from
+                the proposal at every iteration.
+
+        Returns:
+            `NeuralPosterior for chainable calls.
+        """
+        self._rejection_sampling_parameters = parameters
+        return self
+
     @abstractmethod
     def log_prob(
         self, theta: Tensor, x: Optional[Tensor] = None, track_gradients: bool = False
@@ -273,17 +361,13 @@ class NeuralPosterior(ABC):
             self._ensure_single_x(x)
         self._ensure_x_consistent_with_default_x(x)
 
-        return theta, x
+        return theta.to(self._device), x.to(self._device)
 
     def _prepare_for_sample(
-        self,
-        x: Tensor,
-        sample_shape: Optional[Tensor],
-        mcmc_method: Optional[str],
-        mcmc_parameters: Optional[Dict[str, Any]],
-    ) -> Tuple[Tensor, int, str, Dict[str, Any]]:
+        self, x: Tensor, sample_shape: Optional[Tensor]
+    ) -> Tuple[Tensor, int]:
         r"""
-        Return checked and (potentially default) values to sample from the posterior.
+        Return checked, reshaped, potentially default values for `x` and `sample_shape`.
 
         Args:
             sample_shape: Desired shape of samples that are drawn from posterior. If
@@ -292,6 +376,27 @@ class NeuralPosterior(ABC):
             x: Conditioning context for posterior $p(\theta|x)$. If not provided,
                 fall back onto `x_o` if previously provided for multiround training, or
                 to a set default (see `set_default_x()` method).
+
+        Returns: Single (default) $x$ with batch dimension; an integer number of
+            samples.
+        """
+
+        x = atleast_2d_float32_tensor(self._x_else_default_x(x))
+        if not self._allow_iid_x:
+            self._ensure_single_x(x)
+        self._ensure_x_consistent_with_default_x(x)
+        num_samples = torch.Size(sample_shape).numel()
+
+        # Move x to current device.
+        return x.to(self._device), num_samples
+
+    def _potentially_replace_mcmc_parameters(
+        self, mcmc_method: Optional[str], mcmc_parameters: Optional[Dict[str, Any]]
+    ) -> Tuple[str, Dict[str, Any]]:
+        """
+        Return potentially default values to sample from the posterior with MCMC.
+
+        Args:
             mcmc_method: Optional parameter to override `self.mcmc_method`.
             mcmc_parameters: Dictionary overriding the default parameters for MCMC.
                 The following parameters are supported: `thin` to set the thinning
@@ -302,24 +407,39 @@ class NeuralPosterior(ABC):
                 Importance-Resampling using `init_strategy_num_candidates` to find init
                 locations.
 
-        Returns: Single (potentially default) $x$ with batch dimension; an integer
-            number of samples; a (potentially default) mcmc method; and (potentially
+        Returns: A (default) mcmc method and (potentially
             default) mcmc parameters.
         """
-
-        x = atleast_2d_float32_tensor(self._x_else_default_x(x))
-        if not self._allow_iid_x:
-            self._ensure_single_x(x)
-        self._ensure_x_consistent_with_default_x(x)
-        num_samples = torch.Size(sample_shape).numel()
-
         mcmc_method = mcmc_method if mcmc_method is not None else self.mcmc_method
         mcmc_parameters = (
             mcmc_parameters if mcmc_parameters is not None else self.mcmc_parameters
         )
+        return mcmc_method, mcmc_parameters
 
-        # Move x to current device.
-        return x.to(self._device), num_samples, mcmc_method, mcmc_parameters
+    def _potentially_replace_rejection_parameters(
+        self, rejection_sampling_parameters: Optional[Dict[str, Any]]
+    ) -> Dict[str, Any]:
+        """
+        Return potentially default values to rejection sample the posterior.
+
+        Args:
+            rejection_sampling_parameters: Dictionary overriding the default
+                parameters for rejection sampling. The following parameters are
+                supported: `proposal` as the proposal distribtution.
+                `num_samples_to_find_max` as the number of samples that are used to
+                find the maximum of the `potential_fn / proposal` ratio. `m` as
+                multiplier to that ratio. `sampling_batch_size` as the batchsize of
+                samples being drawn from the proposal at every iteration.
+
+        Returns: Potentially default rejection sampling parameters.
+        """
+        rejection_sampling_parameters = (
+            rejection_sampling_parameters
+            if rejection_sampling_parameters is not None
+            else self.rejection_sampling_parameters
+        )
+
+        return rejection_sampling_parameters
 
     def _sample_posterior_mcmc(
         self,
@@ -457,7 +577,7 @@ class NeuralPosterior(ABC):
         samples = samples.reshape(-1, dim_samples)[:num_samples, :]
         assert samples.shape[0] == num_samples
 
-        return samples.type(torch.float32)
+        return samples.type(torch.float32).to(self._device)
 
     def _pyro_mcmc(
         self,
@@ -516,9 +636,11 @@ class NeuralPosterior(ABC):
         condition: Tensor,
         dims_to_sample: List[int],
         x: Optional[Tensor] = None,
+        sample_with: str = "mcmc",
         show_progress_bars: bool = True,
         mcmc_method: Optional[str] = None,
         mcmc_parameters: Optional[Dict[str, Any]] = None,
+        rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
     ) -> Tensor:
         r"""
         Return samples from conditional posterior $p(\theta_i|\theta_j, x)$.
@@ -527,7 +649,7 @@ class NeuralPosterior(ABC):
         from a few parameter dimensions while the other parameter dimensions are kept
         fixed at values specified in `condition`.
 
-        Samples are obtained with MCMC.
+        Samples are obtained with MCMC or rejection sampling.
 
         Args:
             potential_fn_provider: Returns the potential function for the unconditional
@@ -545,48 +667,114 @@ class NeuralPosterior(ABC):
             x: Conditioning context for posterior $p(\theta|x)$. If not provided,
                 fall back onto `x_o` if previously provided for multiround training, or
                 to a set default (see `set_default_x()` method).
+            sample_with: Method to use for sampling from the posterior. Must be one of
+                [`mcmc` | `rejection`]. In this method, the value of
+                `self.sample_with` will be ignored.
             show_progress_bars: Whether to show sampling progress monitor.
             mcmc_method: Optional parameter to override `self.mcmc_method`.
             mcmc_parameters: Dictionary overriding the default parameters for MCMC.
-                The following parameters are supported: `thin` to set the thinning
-                factor for the chain, `warmup_steps` to set the initial number of
-                samples to discard, `num_chains` for the number of chains,
+                The following parameters are supported:
+                `thin` to set the thinning factor for the chain.
+                `warmup_steps` to set the initial number of samples to discard.
+                `num_chains` for the number of chains.
                 `init_strategy` for the initialisation strategy for chains; `prior`
                 will draw init locations from prior, whereas `sir` will use Sequential-
                 Importance-Resampling using `init_strategy_num_candidates` to find init
                 locations.
+                `enable_transform` a bool indicating whether MCMC is performed in
+                z-scored (and unconstrained) space.
+            rejection_sampling_parameters: Dictionary overriding the default parameters
+                for rejection sampling. The following parameters are supported:
+                `proposal` as the proposal distribtution (default is the prior).
+                `max_sampling_batch_size` as the batchsize of samples being drawn from
+                the proposal at every iteration.
+                `num_samples_to_find_max` as the number of samples that are used to
+                find the maximum of the `potential_fn / proposal` ratio.
+                `num_iter_to_find_max` as the number of gradient ascent iterations to
+                find the maximum of that ratio.
+                `m` as multiplier to that ratio.
 
         Returns:
             Samples from conditional posterior.
         """
 
-        x, num_samples, mcmc_method, mcmc_parameters = self._prepare_for_sample(
-            x, sample_shape, mcmc_method, mcmc_parameters
-        )
-
         self.net.eval()
 
-        cond_potential_fn_provider = ConditionalPotentialFunctionProvider(
-            potential_fn_provider, condition, dims_to_sample
-        )
+        x, num_samples = self._prepare_for_sample(x, sample_shape)
 
-        samples = self._sample_posterior_mcmc(
-            num_samples=num_samples,
-            potential_fn=cond_potential_fn_provider(
-                self._prior, self.net, x, mcmc_method
-            ),
-            init_fn=self._build_mcmc_init_fn(
-                # Restrict prior to sample only free dimensions.
-                RestrictedPriorForConditional(self._prior, dims_to_sample),
-                cond_potential_fn_provider(self._prior, self.net, x, "slice_np"),
+        if sample_with == "mcmc":
+            mcmc_method, mcmc_parameters = self._potentially_replace_mcmc_parameters(
+                mcmc_method, mcmc_parameters
+            )
+            transform = mcmc_transform(
+                self._prior, device=self._device, **mcmc_parameters
+            )
+            transform_dims_to_sample = RestrictedTransformForConditional(
+                transform, condition, dims_to_sample
+            )
+
+            condition = atleast_2d_float32_tensor(condition)
+
+            # Transform the `condition` to unconstrained space.
+            transformed_condition = transform(condition)
+            cond_potential_fn_provider = ConditionalPotentialFunctionProvider(
+                potential_fn_provider, transformed_condition, dims_to_sample
+            )
+
+            transformed_samples = self._sample_posterior_mcmc(
+                num_samples=num_samples,
+                potential_fn=cond_potential_fn_provider(
+                    self._prior,
+                    self.net,
+                    x,
+                    mcmc_method,
+                    transform,
+                ),
+                init_fn=self._build_mcmc_init_fn(
+                    # Restrict prior to sample only free dimensions.
+                    RestrictedPriorForConditional(self._prior, dims_to_sample),
+                    cond_potential_fn_provider(
+                        self._prior,
+                        self.net,
+                        x,
+                        "slice_np",
+                        transform,
+                    ),
+                    transform=transform_dims_to_sample,
+                    **mcmc_parameters,
+                ),
+                mcmc_method=mcmc_method,
+                condition=condition,
+                dims_to_sample=dims_to_sample,
+                show_progress_bars=show_progress_bars,
                 **mcmc_parameters,
-            ),
-            mcmc_method=mcmc_method,
-            condition=condition,
-            dims_to_sample=dims_to_sample,
-            show_progress_bars=show_progress_bars,
-            **mcmc_parameters,
-        )
+            )
+            samples = transform_dims_to_sample.inv(transformed_samples)
+        elif sample_with == "rejection":
+            cond_potential_fn_provider = ConditionalPotentialFunctionProvider(
+                potential_fn_provider, condition, dims_to_sample
+            )
+            rejection_sampling_parameters = (
+                self._potentially_replace_rejection_parameters(
+                    rejection_sampling_parameters
+                )
+            )
+            if "proposal" not in rejection_sampling_parameters:
+                rejection_sampling_parameters[
+                    "proposal"
+                ] = RestrictedPriorForConditional(self._prior, dims_to_sample)
+
+            samples, _ = rejection_sample(
+                potential_fn=cond_potential_fn_provider(
+                    self._prior, self.net, x, "rejection"
+                ),
+                num_samples=num_samples,
+                **rejection_sampling_parameters,
+            )
+        else:
+            raise NameError(
+                "The only implemented sampling methods are `mcmc` and `rejection`."
+            )
 
         self.net.train(True)
 
@@ -657,41 +845,6 @@ class NeuralPosterior(ABC):
             "https://github.com/mackelab/sbi/issues"
         )
 
-        # If the prior is `BoxUniform`, define a transformation to optimize in
-        # unbounded space.
-        is_boxuniform, boxuniform = check_dist_class(self._prior, BoxUniform)
-        if is_boxuniform:
-
-            def tf_inv(theta_t):
-                return utils.expit(
-                    theta_t,
-                    torch.as_tensor(
-                        boxuniform.support.base_constraint.lower_bound, dtype=float32
-                    ),
-                    torch.as_tensor(
-                        boxuniform.support.base_constraint.upper_bound, dtype=float32
-                    ),
-                )
-
-            def tf(theta):
-                return utils.logit(
-                    theta,
-                    torch.as_tensor(
-                        boxuniform.support.base_constraint.lower_bound, dtype=float32
-                    ),
-                    torch.as_tensor(
-                        boxuniform.support.base_constraint.upper_bound, dtype=float32
-                    ),
-                )
-
-        else:
-
-            def tf_inv(theta_t):
-                return theta_t
-
-            def tf(theta):
-                return theta
-
         if isinstance(init_method, str):
             # Find initial position.
             if init_method == "posterior":
@@ -710,84 +863,31 @@ class NeuralPosterior(ABC):
         else:
             inits = init_method
 
-        init_probs = self.log_prob(inits, x=x, **log_prob_kwargs)
+        def potential_fn(theta):
+            return self.log_prob(theta, x=x, track_gradients=True, **log_prob_kwargs)
 
-        # Pick the `num_to_optimize` best init locations.
-        sort_indices = torch.argsort(init_probs, dim=0)
-        sorted_inits = inits[sort_indices]
-        optimize_inits = sorted_inits[-num_to_optimize:]
+        interruption_note = """The last estimate of the MAP can be accessed via the
+                            `posterior.map_` attribute."""
 
-        # The `_overall` variables store data accross the iterations, whereas the
-        # `_iter` variables contain data exclusively extracted from the current
-        # iteration.
-        best_log_prob_iter = torch.max(init_probs)
-        best_theta_iter = sorted_inits[-1]
-        best_theta_overall = best_theta_iter.detach().clone()
-        best_log_prob_overall = best_log_prob_iter.detach().clone()
+        self.map_, _ = optimize_potential_fn(
+            potential_fn=potential_fn,
+            inits=inits,
+            dist_specifying_bounds=self._prior,
+            num_iter=num_iter,
+            learning_rate=learning_rate,
+            num_to_optimize=num_to_optimize,
+            save_best_every=save_best_every,
+            show_progress_bars=show_progress_bars,
+            interruption_note=interruption_note,
+        )
 
-        self.map_ = best_theta_overall
-
-        optimize_inits = tf(optimize_inits)
-        optimize_inits.requires_grad_(True)
-        optimizer = optim.Adam([optimize_inits], lr=learning_rate)
-
-        iter_ = 0
-
-        # Try-except block in case the user interrupts the program and wants to fall
-        # back on the last saved `.map_`. We want to avoid a long error-message here.
-        try:
-
-            while iter_ < num_iter:
-
-                optimizer.zero_grad()
-                probs = self.log_prob(
-                    tf_inv(optimize_inits), x=x, track_gradients=True, **log_prob_kwargs
-                ).squeeze()
-                loss = -probs.sum()
-                loss.backward()
-                optimizer.step()
-
-                with torch.no_grad():
-                    if iter_ % save_best_every == 0 or iter_ == num_iter - 1:
-                        # Evaluate the optimized locations and pick the best one.
-                        log_probs_of_optimized = self.log_prob(
-                            tf_inv(optimize_inits), x=x, **log_prob_kwargs
-                        )
-                        best_theta_iter = optimize_inits[
-                            torch.argmax(log_probs_of_optimized)
-                        ]
-                        best_log_prob_iter = self.log_prob(
-                            tf_inv(best_theta_iter), x=x, **log_prob_kwargs
-                        )
-                        if best_log_prob_iter > best_log_prob_overall:
-                            best_theta_overall = best_theta_iter.detach().clone()
-                            best_log_prob_overall = best_log_prob_iter.detach().clone()
-
-                    if show_progress_bars:
-                        print(
-                            f"""Optimizing MAP estimate. Iterations: {iter_+1} /
-                            {num_iter}. Performance in iteration
-                            {divmod(iter_+1, save_best_every)[0] * save_best_every}:
-                            {best_log_prob_iter.item():.2f} (= unnormalized log-prob""",
-                            end="\r",
-                        )
-                    self.map_ = tf_inv(best_theta_overall)
-
-                iter_ += 1
-
-        except KeyboardInterrupt:
-            print(
-                f"Optimization was interrupted after {iter_} iterations. The last "
-                "estimate of the MAP can be accessed via the `posterior.map_` "
-                "attribute."
-            )
-
-        return tf_inv(best_theta_overall)
+        return self.map_
 
     def _build_mcmc_init_fn(
         self,
         prior: Any,
         potential_fn: Callable,
+        transform: torch_tf.Transform,
         init_strategy: str = "prior",
         **kwargs,
     ) -> Callable:
@@ -806,9 +906,9 @@ class NeuralPosterior(ABC):
         Returns: Initialization function.
         """
         if init_strategy == "prior":
-            return lambda: prior_init(prior, **kwargs)
+            return lambda: prior_init(prior, transform=transform, **kwargs)
         elif init_strategy == "sir":
-            return lambda: sir(prior, potential_fn, **kwargs)
+            return lambda: sir(prior, potential_fn, transform=transform, **kwargs)
         elif init_strategy == "latest_sample":
             latest_sample = IterateParameters(self._mcmc_init_params, **kwargs)
             return latest_sample
@@ -1014,6 +1114,13 @@ class NeuralPosterior(ABC):
             warning_msg,
         )
 
+        state_dict, warning_msg = check_warn_and_setstate(
+            state_dict,
+            "_sample_with",
+            "rejection" if state_dict["_method_family"] == "snpe" else "mcmc",
+            warning_msg,
+        )
+
         if warning_msg:
             warning_description = (
                 "You had saved the posterior under an older version of `sbi`. To make "
@@ -1053,23 +1160,42 @@ class ConditionalPotentialFunctionProvider:
         self.condition = ensure_theta_batched(condition)
         self.dims_to_sample = dims_to_sample
 
-    def __call__(self, prior, net: nn.Module, x: Tensor, mcmc_method: str) -> Callable:
+    def __call__(
+        self,
+        prior,
+        net: nn.Module,
+        x: Tensor,
+        method: str,
+        transform: torch_tf.Transform = torch_tf.identity_transform,
+    ) -> Callable:
         """Return potential function.
 
-        Switch on numpy or pyro potential function based on `mcmc_method`.
+        Switch on numpy or pyro potential function based on `method`.
         """
         # Set prior, net, and x as attributes of unconditional potential_fn_provider.
-        _ = self.potential_fn_provider.__call__(prior, net, x, mcmc_method)
+        _ = self.potential_fn_provider.__call__(prior, net, x, method, transform)
+        self.device = next(net.parameters()).device
 
-        if mcmc_method in ("slice", "hmc", "nuts"):
-            return self.pyro_potential
+        if method == "slice":
+            return partial(self.pyro_potential, track_gradients=False)
+        elif method in ("hmc", "nuts"):
+            return partial(self.pyro_potential, track_gradients=True)
+        elif "slice_np" in method:
+            return partial(self.posterior_potential, track_gradients=False)
+        elif method == "rejection":
+            return partial(self.posterior_potential, track_gradients=True)
         else:
-            return self.np_potential
+            NotImplementedError
 
-    def np_potential(self, theta: np.ndarray) -> ScalarFloat:
+    def posterior_potential(
+        self, theta: np.ndarray, track_gradients: bool = False
+    ) -> ScalarFloat:
         r"""
         Return conditional posterior log-probability or $-\infty$ if outside prior.
 
+        The only differences to the `np_potential` is that it tracks the gradients and
+        does not return a `numpy` array.
+
         Args:
             theta: Free parameters $\theta_i$, batch dimension 1.
 
@@ -1077,16 +1203,21 @@ class ConditionalPotentialFunctionProvider:
             Conditional posterior log-probability $\log(p(\theta_i|\theta_j, x))$,
             masked outside of prior.
         """
-        theta = torch.as_tensor(theta, dtype=torch.float32)
-
-        theta_condition = deepcopy(self.condition)
-        theta_condition[:, self.dims_to_sample] = theta
-
-        return self.potential_fn_provider.np_potential(
-            utils.tensor2numpy(theta_condition)
+        theta = ensure_theta_batched(torch.as_tensor(theta, dtype=torch.float32)).to(
+            self.device
         )
 
-    def pyro_potential(self, theta: Dict[str, Tensor]) -> Tensor:
+        theta_condition = deepcopy(self.condition).to(self.device)
+        theta_condition = theta_condition.repeat(theta.shape[0], 1)
+        theta_condition[:, self.dims_to_sample] = theta
+
+        return self.potential_fn_provider.posterior_potential(
+            theta_condition, track_gradients=track_gradients
+        )
+
+    def pyro_potential(
+        self, theta: Dict[str, Tensor], track_gradients: bool = False
+    ) -> Tensor:
         r"""
         Return conditional posterior log-probability or $-\infty$ if outside prior.
 
@@ -1099,11 +1230,14 @@ class ConditionalPotentialFunctionProvider:
         """
 
         theta = next(iter(theta.values()))
+        theta = ensure_theta_batched(theta).to(self.device)
 
-        theta_condition = deepcopy(self.condition)
+        theta_condition = deepcopy(self.condition).to(self.device)
         theta_condition[:, self.dims_to_sample] = theta
 
-        return self.potential_fn_provider.pyro_potential({"": theta_condition})
+        return self.potential_fn_provider.pyro_potential(
+            {"": theta_condition}, track_gradients=track_gradients
+        )
 
 
 class RestrictedPriorForConditional:
@@ -1137,3 +1271,60 @@ class RestrictedPriorForConditional:
         the $\theta$ under the full joint once we have added the condition.
         """
         return self.full_prior.log_prob(*args, **kwargs)
+
+
+class RestrictedTransformForConditional(nn.Module):
+    """
+    Class to restrict the transform to fewer dimensions for conditional sampling.
+
+    The resulting transform transforms only the free dimensions of the conditional.
+    Notably, the `log_abs_det` is computed given all dimensions. However, the
+    `log_abs_det` stemming from the fixed dimensions is a constant and drops out during
+    MCMC.
+
+    This is needed for the the MCMC initialization functions when conditioning and
+    when transforming the samples back into the original theta space after sampling.
+    """
+
+    def __init__(
+        self,
+        transform: torch_tf.Transform,
+        condition: Tensor,
+        dims_to_sample: List[int],
+    ) -> None:
+        super().__init__()
+        self.transform = transform
+        self.condition = ensure_theta_batched(condition)
+        self.dims_to_sample = dims_to_sample
+
+    def forward(self, theta: Tensor) -> Tuple[Tensor, Tensor]:
+        r"""
+        Transform restricted $\theta$.
+        """
+        full_theta = self.condition.repeat(theta.shape[0], 1)
+        full_theta[:, self.dims_to_sample] = theta
+        tf_full_theta = self.transform(full_theta)
+        return tf_full_theta[:, self.dims_to_sample]
+
+    def inv(self, theta: Tensor) -> Tuple[Tensor, Tensor]:
+        r"""
+        Inverse transform restricted $\theta$.
+        """
+        full_theta = self.condition.repeat(theta.shape[0], 1)
+        full_theta[:, self.dims_to_sample] = theta
+        tf_full_theta = self.transform.inv(full_theta)
+        return tf_full_theta[:, self.dims_to_sample]
+
+    def log_abs_det_jacobian(self, theta1: Tensor, theta2: Tensor) -> Tensor:
+        """
+        Return the `log_abs_det_jacobian` of |dtheta1 / dtheta2|.
+
+        The determinant is summed over all dimensions, not just the `dims_to_sample`
+        ones.
+        """
+        full_theta1 = self.condition.repeat(theta1.shape[0], 1)
+        full_theta1[:, self.dims_to_sample] = theta1
+        full_theta2 = self.condition.repeat(theta2.shape[0], 1)
+        full_theta2[:, self.dims_to_sample] = theta2
+        log_abs_det = self.transform.log_abs_det_jacobian(full_theta1, full_theta2)
+        return log_abs_det
