@@ -1,22 +1,29 @@
 # This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
 # under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.
 
+
 from functools import partial
-from typing import Any, Callable, Dict, List, Optional, Union
+from typing import Any, Callable, Dict, List, Optional, Union, Tuple
+from warnings import warn
 
 import numpy as np
 import torch
+import torch.distributions.transforms as torch_tf
 from torch import Tensor, log, nn
+from warnings import warn
 
 from sbi import utils as utils
+from pyknos.mdn.mdn import MultivariateGaussianMDN as mdn
 from sbi.inference.posteriors.base_posterior import NeuralPosterior
-from sbi.types import ScalarFloat, Shape
-from sbi.utils import del_entries, within_support
+from sbi.types import Shape
+from sbi.utils import del_entries, mcmc_transform, rejection_sample, within_support
 from sbi.utils.torchutils import (
     atleast_2d,
     batched_first_of_batch,
     ensure_theta_batched,
+    atleast_2d_float32_tensor,
 )
+from sbi.utils.conditional_density import extract_and_transform_mog, condition_mog
 
 
 class DirectPosterior(NeuralPosterior):
@@ -41,10 +48,10 @@ class DirectPosterior(NeuralPosterior):
         neural_net: nn.Module,
         prior,
         x_shape: torch.Size,
-        rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
-        sample_with_mcmc: bool = True,
+        sample_with: str = "rejection",
         mcmc_method: str = "slice_np",
         mcmc_parameters: Optional[Dict[str, Any]] = None,
+        rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
         device: str = "cpu",
     ):
         """
@@ -53,14 +60,10 @@ class DirectPosterior(NeuralPosterior):
             neural_net: A classifier for SNRE, a density estimator for SNPE and SNL.
             prior: Prior distribution with `.log_prob()` and `.sample()`.
             x_shape: Shape of a single simulator output.
-            rejection_sampling_parameters: Dictonary overriding the default parameters
-                for rejection sampling. The following parameters are supported:
-                `max_sampling_batch_size` to set the batch size for drawing new
-                samples from the candidate distribution, e.g., the posterior. Larger
-                batch size speeds up sampling.
-            sample_with_mcmc: Whether to sample with MCMC. Will always be `True` for SRE
-                and SNL, but can also be set to `True` for SNPE if MCMC is preferred to
-                deal with leakage over rejection sampling.
+            sample_with: Method to use for sampling from the posterior. Must be one of
+                [`mcmc` | `rejection`]. With default parameters, `rejection` samples
+                from the posterior estimated by the neural net and rejects only if the
+                samples are outside of the prior support.
             mcmc_method: Method used for MCMC sampling, one of `slice_np`, `slice`,
                 `hmc`, `nuts`. Currently defaults to `slice_np` for a custom numpy
                 implementation of slice sampling; select `hmc`, `nuts` or `slice` for
@@ -73,41 +76,48 @@ class DirectPosterior(NeuralPosterior):
                 will draw init locations from prior, whereas `sir` will use Sequential-
                 Importance-Resampling using `init_strategy_num_candidates` to find init
                 locations.
+            rejection_sampling_parameters: Dictionary overriding the default parameters
+                for rejection sampling. The following parameters are supported:
+                `proposal` as the proposal distribtution (default is the trained
+                neural net). `max_sampling_batch_size` as the batchsize of samples
+                being drawn from the proposal at every iteration.
+                `num_samples_to_find_max` as the number of samples that are used to
+                find the maximum of the `potential_fn / proposal` ratio.
+                `num_iter_to_find_max` as the number of gradient ascent iterations to
+                find the maximum of that ratio. `m` as multiplier to that ratio.
             device: Training device, e.g., cpu or cuda:0
         """
 
-        kwargs = del_entries(
-            locals(),
-            entries=(
-                "self",
-                "__class__",
-                "sample_with_mcmc",
-                "rejection_sampling_parameters",
-            ),
-        )
+        kwargs = del_entries(locals(), entries=("self", "__class__"))
         super().__init__(**kwargs)
 
-        self.set_sample_with_mcmc(sample_with_mcmc)
-        self.set_rejection_sampling_parameters(rejection_sampling_parameters)
         self._purpose = (
             "It allows to .sample() and .log_prob() the posterior and wraps the "
             "output of the .net to avoid leakage into regions with 0 prior probability."
         )
 
     @property
-    def sample_with_mcmc(self) -> bool:
+    def _sample_with_mcmc(self) -> bool:
         """
+        Deprecated, will be removed in future versions of `sbi`.
+
         Return `True` if NeuralPosterior instance should use MCMC in `.sample()`.
         """
         return self._sample_with_mcmc
 
-    @sample_with_mcmc.setter
-    def sample_with_mcmc(self, value: bool) -> None:
-        """See `set_sample_with_mcmc`."""
-        self.set_sample_with_mcmc(value)
+    @_sample_with_mcmc.setter
+    def _sample_with_mcmc(self, value: bool) -> None:
+        """
+        Deprecated, will be removed in future versions of `sbi`.
 
-    def set_sample_with_mcmc(self, use_mcmc: bool) -> "NeuralPosterior":
-        """Turns MCMC sampling on or off and returns `NeuralPosterior`.
+        See `set_sample_with_mcmc`."""
+        self._set_sample_with_mcmc(value)
+
+    def _set_sample_with_mcmc(self, use_mcmc: bool) -> "NeuralPosterior":
+        """
+        Deprecated, will be removed in future versions of `sbi`.
+
+        Turns MCMC sampling on or off and returns `NeuralPosterior`.
 
         Args:
             use_mcmc: Flag to set whether or not MCMC sampling is used.
@@ -119,38 +129,16 @@ class DirectPosterior(NeuralPosterior):
             ValueError: on attempt to turn off MCMC sampling for family of methods that
                 do not support rejection sampling.
         """
-        self._sample_with_mcmc = use_mcmc
-        return self
-
-    @property
-    def rejection_sampling_parameters(self) -> dict:
-        """Returns rejection sampling parameter."""
-        if self._rejection_sampling_parameters is None:
-            return {}
+        warn(
+            f"You set `sample_with_mcmc={use_mcmc}`. This is deprecated "
+            "since `sbi v0.17.0` and will lead to an error in future versions. "
+            "Please use `sample_with='mcmc'` instead."
+        )
+        if use_mcmc:
+            self.set_sample_with("mcmc")
         else:
-            return self._rejection_sampling_parameters
-
-    @rejection_sampling_parameters.setter
-    def rejection_sampling_parameters(self, parameters: Dict[str, Any]) -> None:
-        """See `set_rejection_sampling_parameters`."""
-        self.set_rejection_sampling_parameters(parameters)
-
-    def set_rejection_sampling_parameters(
-        self, parameters: Dict[str, Any]
-    ) -> "NeuralPosterior":
-        """Sets parameters for rejection sampling and returns `NeuralPosterior`.
-
-        Args:
-            parameters: Dictonary overriding the default parameters
-                for rejection sampling. The following parameters are supported:
-                `max_sampling_batch_size` to the set the batch size for drawing new
-                samples from the candidate distribution, e.g., the posterior. Larger
-                batch size speeds up sampling.
-
-        Returns:
-            `NeuralPosterior for chainable calls.
-        """
-        self._rejection_sampling_parameters = parameters
+            self.set_sample_with("rejection")
+        self._sample_with_mcmc = use_mcmc
         return self
 
     def log_prob(
@@ -187,7 +175,6 @@ class DirectPosterior(NeuralPosterior):
         Returns:
             `(len(θ),)`-shaped log posterior probability $\log p(\theta|x)$ for θ in the
             support of the prior, -∞ (corresponding to 0 probability) outside.
-
         """
 
         # TODO Train exited here, entered after sampling?
@@ -199,9 +186,7 @@ class DirectPosterior(NeuralPosterior):
         with torch.set_grad_enabled(track_gradients):
 
             # Evaluate on device, move back to cpu for comparison with prior.
-            unnorm_log_prob = self.net.log_prob(
-                theta_repeated.to(self._device), x_repeated.to(self._device)
-            ).cpu()
+            unnorm_log_prob = self.net.log_prob(theta_repeated, x_repeated)
 
             # Force probability to be zero outside prior support.
             in_prior_support = within_support(self._prior, theta)
@@ -209,7 +194,7 @@ class DirectPosterior(NeuralPosterior):
             masked_log_prob = torch.where(
                 in_prior_support,
                 unnorm_log_prob,
-                torch.tensor(float("-inf"), dtype=torch.float32),
+                torch.tensor(float("-inf"), dtype=torch.float32, device=self._device),
             )
 
             if leakage_correction_params is None:
@@ -259,12 +244,13 @@ class DirectPosterior(NeuralPosterior):
         """
 
         def acceptance_at(x: Tensor) -> Tensor:
-            return utils.sample_posterior_within_prior(
-                self.net,
-                self._prior,
-                x.to(self._device),
-                num_rejection_samples,
-                show_progress_bars,
+
+            return utils.rejection_sample_posterior_within_prior(
+                posterior_nn=self.net,
+                prior=self._prior,
+                x=x.to(self._device),
+                num_samples=num_rejection_samples,
+                show_progress_bars=show_progress_bars,
                 sample_for_correction_factor=True,
                 max_sampling_batch_size=rejection_sampling_batch_size,
             )[1]
@@ -288,10 +274,11 @@ class DirectPosterior(NeuralPosterior):
         sample_shape: Shape = torch.Size(),
         x: Optional[Tensor] = None,
         show_progress_bars: bool = True,
-        sample_with_mcmc: Optional[bool] = None,
+        sample_with: Optional[str] = None,
         mcmc_method: Optional[str] = None,
         mcmc_parameters: Optional[Dict[str, Any]] = None,
         rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
+        sample_with_mcmc: Optional[bool] = None,
     ) -> Tensor:
         r"""
         Return samples from posterior distribution $p(\theta|x)$.
@@ -308,62 +295,117 @@ class DirectPosterior(NeuralPosterior):
             x: Conditioning context for posterior $p(\theta|x)$. If not provided,
                 fall back onto `x` passed to `set_default_x()`.
             show_progress_bars: Whether to show sampling progress monitor.
-            sample_with_mcmc: Optional parameter to override `self.sample_with_mcmc`.
+            sample_with: Method to use for sampling from the posterior. Must be one of
+                [`mcmc` | `rejection`]. With default parameters, `rejection` samples
+                from the posterior estimated by the neural net and rejects only if the
+                samples are outside of the prior support.
             mcmc_method: Optional parameter to override `self.mcmc_method`.
             mcmc_parameters: Dictionary overriding the default parameters for MCMC.
-                The following parameters are supported: `thin` to set the thinning
-                factor for the chain, `warmup_steps` to set the initial number of
-                samples to discard, `num_chains` for the number of chains,
+                The following parameters are supported:
+                `thin` to set the thinning factor for the chain.
+                `warmup_steps` to set the initial number of samples to discard.
+                `num_chains` for the number of chains.
                 `init_strategy` for the initialisation strategy for chains; `prior`
                 will draw init locations from prior, whereas `sir` will use Sequential-
                 Importance-Resampling using `init_strategy_num_candidates` to find init
                 locations.
+                `enable_transform` a bool indicating whether MCMC is performed in
+                z-scored (and unconstrained) space.
             rejection_sampling_parameters: Dictionary overriding the default parameters
                 for rejection sampling. The following parameters are supported:
-                `max_sampling_batch_size` to set the batch size for drawing new
-                samples from the candidate distribution, e.g., the posterior. Larger
-                batch size speeds up sampling.
+                `proposal` as the proposal distribtution (default is the trained
+                neural net).
+                `max_sampling_batch_size` as the batchsize of samples being drawn from
+                the proposal at every iteration.
+                `num_samples_to_find_max` as the number of samples that are used to
+                find the maximum of the `potential_fn / proposal` ratio.
+                `num_iter_to_find_max` as the number of gradient ascent iterations to
+                find the maximum of that ratio.
+                `m` as multiplier to that ratio.
+            sample_with_mcmc: Deprecated since `sbi v0.17.0`. Use `sample_with=mcmc`
+                instead.
+
         Returns:
             Samples from posterior.
         """
 
-        x, num_samples, mcmc_method, mcmc_parameters = self._prepare_for_sample(
-            x, sample_shape, mcmc_method, mcmc_parameters
-        )
-
-        sample_with_mcmc = (
-            sample_with_mcmc if sample_with_mcmc is not None else self.sample_with_mcmc
-        )
+        if sample_with_mcmc is not None:
+            warn(
+                f"You set `sample_with_mcmc={sample_with_mcmc}`. This is deprecated "
+                "since `sbi v0.17.0` and will lead to an error in future versions. "
+                "Please use `sample_with='mcmc'` instead."
+            )
+            if sample_with_mcmc:
+                sample_with = "mcmc"
 
         self.net.eval()
 
-        if sample_with_mcmc:
-            potential_fn_provider = PotentialFunctionProvider()
-            samples = self._sample_posterior_mcmc(
+        sample_with = sample_with if sample_with is not None else self._sample_with
+
+        x, num_samples = self._prepare_for_sample(x, sample_shape)
+
+        potential_fn_provider = PotentialFunctionProvider()
+        if sample_with == "mcmc":
+            mcmc_method, mcmc_parameters = self._potentially_replace_mcmc_parameters(
+                mcmc_method, mcmc_parameters
+            )
+            transform = mcmc_transform(
+                self._prior, device=self._device, **mcmc_parameters
+            )
+            transformed_samples = self._sample_posterior_mcmc(
                 num_samples=num_samples,
                 potential_fn=potential_fn_provider(
-                    self._prior, self.net, x, mcmc_method
+                    self._prior, self.net, x, mcmc_method, transform
                 ),
                 init_fn=self._build_mcmc_init_fn(
                     self._prior,
-                    potential_fn_provider(self._prior, self.net, x, "slice_np"),
+                    potential_fn_provider(
+                        self._prior, self.net, x, "slice_np", transform
+                    ),
+                    transform=transform,
                     **mcmc_parameters,
                 ),
                 mcmc_method=mcmc_method,
                 show_progress_bars=show_progress_bars,
                 **mcmc_parameters,
             )
+            samples = transform.inv(transformed_samples)
+        elif sample_with == "rejection":
+            rejection_sampling_parameters = (
+                self._potentially_replace_rejection_parameters(
+                    rejection_sampling_parameters
+                )
+            )
+            if "proposal" not in rejection_sampling_parameters:
+                assert (
+                    not self.net.training
+                ), "Posterior nn must be in eval mode for sampling."
+
+                # If the user does not explictly pass a `proposal`, we sample from the
+                # neural net estimating the posterior and reject only those samples
+                # that are outside of the prior support. This can be considered as
+                # rejection sampling with a very good proposal.
+                samples = utils.rejection_sample_posterior_within_prior(
+                    posterior_nn=self.net,
+                    prior=self._prior,
+                    x=x.to(self._device),
+                    num_samples=num_samples,
+                    show_progress_bars=show_progress_bars,
+                    sample_for_correction_factor=True,
+                    **rejection_sampling_parameters,
+                )[0]
+            else:
+                samples, _ = rejection_sample(
+                    potential_fn=potential_fn_provider(
+                        self._prior, self.net, x, "rejection"
+                    ),
+                    num_samples=num_samples,
+                    show_progress_bars=show_progress_bars,
+                    **rejection_sampling_parameters,
+                )
         else:
-            # Rejection sampling.
-            samples, _ = utils.sample_posterior_within_prior(
-                self.net,
-                self._prior,
-                x,
-                num_samples=num_samples,
-                show_progress_bars=show_progress_bars,
-                **rejection_sampling_parameters
-                if (rejection_sampling_parameters is not None)
-                else self.rejection_sampling_parameters,
+            raise NameError(
+                "The only implemented sampling methods are `mcmc` and `rejection`."
             )
 
         self.net.train(True)
@@ -376,9 +418,11 @@ class DirectPosterior(NeuralPosterior):
         condition: Tensor,
         dims_to_sample: List[int],
         x: Optional[Tensor] = None,
+        sample_with: str = "mcmc",
         show_progress_bars: bool = True,
         mcmc_method: Optional[str] = None,
         mcmc_parameters: Optional[Dict[str, Any]] = None,
+        rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
     ) -> Tensor:
         r"""
         Return samples from conditional posterior $p(\theta_i|\theta_j, x)$.
@@ -402,31 +446,140 @@ class DirectPosterior(NeuralPosterior):
                 `condition`.
             x: Conditioning context for posterior $p(\theta|x)$. If not provided,
                 fall back onto `x` passed to `set_default_x()`.
+            sample_with: Method to use for sampling from the posterior. Must be one of
+                [`mcmc` | `rejection`]. In this method, the value of
+                `self.sample_with` will be ignored.
             show_progress_bars: Whether to show sampling progress monitor.
             mcmc_method: Optional parameter to override `self.mcmc_method`.
             mcmc_parameters: Dictionary overriding the default parameters for MCMC.
-                The following parameters are supported: `thin` to set the thinning
-                factor for the chain, `warmup_steps` to set the initial number of
-                samples to discard, `num_chains` for the number of chains,
+                The following parameters are supported:
+                `thin` to set the thinning factor for the chain.
+                `warmup_steps` to set the initial number of samples to discard.
+                `num_chains` for the number of chains.
                 `init_strategy` for the initialisation strategy for chains; `prior`
                 will draw init locations from prior, whereas `sir` will use Sequential-
                 Importance-Resampling using `init_strategy_num_candidates` to find init
                 locations.
+                `enable_transform` a bool indicating whether MCMC is performed in
+                z-scored (and unconstrained) space.
+            rejection_sampling_parameters: Dictionary overriding the default parameters
+                for rejection sampling. The following parameters are supported:
+                `proposal` as the proposal distribtution (default is the prior).
+                `max_sampling_batch_size` as the batchsize of samples being drawn from
+                the proposal at every iteration.
+                `num_samples_to_find_max` as the number of samples that are used to
+                find the maximum of the `potential_fn / proposal` ratio.
+                `num_iter_to_find_max` as the number of gradient ascent iterations to
+                find the maximum of that ratio.
+                `m` as multiplier to that ratio.
 
         Returns:
             Samples from conditional posterior.
         """
+        if not hasattr(self.net, "_distribution"):
+            raise NotImplementedError(
+                "`sample_conditional` is not implemented for SNPE-A."
+            )
 
-        return super().sample_conditional(
-            PotentialFunctionProvider(),
-            sample_shape,
-            condition,
-            dims_to_sample,
-            x,
-            show_progress_bars,
-            mcmc_method,
-            mcmc_parameters,
-        )
+        net = self.net
+        x = atleast_2d_float32_tensor(self._x_else_default_x(x))
+
+        if type(net._distribution) is mdn:
+            condition = atleast_2d_float32_tensor(condition)
+            num_samples = torch.Size(sample_shape).numel()
+
+            logits, means, precfs, _ = extract_and_transform_mog(nn=net, context=x)
+            logits, means, precfs, _ = condition_mog(
+                self._prior, condition, dims_to_sample, logits, means, precfs
+            )
+
+            # Currently difficult to integrate `sample_posterior_within_prior`.
+            warn(
+                "Sampling MoG analytically. "
+                "Some of the samples might not be within the prior support!"
+            )
+            samples = mdn.sample_mog(num_samples, logits, means, precfs)
+            return samples.detach().reshape((*sample_shape, -1))
+
+        else:
+            return super().sample_conditional(
+                PotentialFunctionProvider(),
+                sample_shape,
+                condition,
+                dims_to_sample,
+                x,
+                sample_with,
+                show_progress_bars,
+                mcmc_method,
+                mcmc_parameters,
+                rejection_sampling_parameters,
+            )
+
+    def log_prob_conditional(
+        self,
+        theta: Tensor,
+        condition: Tensor,
+        dims_to_evaluate: List[int],
+        x: Optional[Tensor] = None,
+    ) -> Tensor:
+        """Evaluates the conditional posterior probability of a MDN at a context x for
+        a value theta given a condition.
+
+        This function only works for MDN based posteriors, becuase evaluation is done
+        analytically. For all other density estimators a `NotImplementedError` will be
+        raised!
+
+        Args:
+            theta: Parameters $\theta$.
+            condition: Parameter set that all dimensions not specified in
+                `dims_to_sample` will be fixed to. Should contain dim_theta elements,
+                i.e. it could e.g. be a sample from the posterior distribution.
+                The entries at all `dims_to_sample` will be ignored.
+            dims_to_evaluate: Which dimensions to evaluate the sample for.
+                The dimensions not specified in `dims_to_evaluate` will be fixed to values given in `condition`.
+            x: Conditioning context for posterior $p(\theta|x)$. If not provided,
+                fall back onto `x` passed to `set_default_x()`.
+
+        Returns:
+            log_prob: `(len(θ),)`-shaped normalized (!) log posterior probability
+                $\log p(\theta|x) for θ in the support of the prior, -∞ (corresponding
+                to 0 probability) outside.
+        """
+
+        if type(self.net._distribution) == mdn:
+            logits, means, precfs, sumlogdiag = extract_and_transform_mog(self, x)
+            logits, means, precfs, sumlogdiag = condition_mog(
+                self._prior, condition, dims_to_evaluate, logits, means, precfs
+            )
+
+            batch_size, dim = theta.shape
+            prec = precfs.transpose(3, 2) @ precfs
+
+            self.net.eval()  # leakage correction requires eval mode
+
+            if dim != len(dims_to_evaluate):
+                X = X[:, dims_to_evaluate]
+
+            # Implementing leakage correction is difficult for conditioned MDNs,
+            # because samples from self i.e. the full posterior are used rather
+            # then from the new, conditioned posterior.
+            warn("Probabilities are not adjusted for leakage.")
+
+            log_prob = mdn.log_prob_mog(
+                theta,
+                logits.repeat(batch_size, 1),
+                means.repeat(batch_size, 1, 1),
+                prec.repeat(batch_size, 1, 1, 1),
+                sumlogdiag.repeat(batch_size, 1),
+            )
+
+            self.net.train(True)
+            return log_prob.detach()
+
+        else:
+            raise NotImplementedError(
+                "This functionality is only available for MDN based posteriors."
+            )
 
     def map(
         self,
@@ -515,76 +668,83 @@ class PotentialFunctionProvider:
         prior,
         posterior_nn: nn.Module,
         x: Tensor,
-        mcmc_method: str,
+        method: str,
+        transform: torch_tf.Transform = torch_tf.identity_transform,
     ) -> Callable:
         """Return potential function.
 
-        Switch on numpy or pyro potential function based on `mcmc_method`.
+        Switch on numpy or pyro potential function based on `method`.
         """
         self.posterior_nn = posterior_nn
         self.prior = prior
         self.device = next(posterior_nn.parameters()).device
         self.x = atleast_2d(x).to(self.device)
+        self.transform = transform
 
-        if mcmc_method == "slice":
+        if method == "slice":
             return partial(self.pyro_potential, track_gradients=False)
-        elif mcmc_method in ("hmc", "nuts"):
+        elif method in ("hmc", "nuts"):
             return partial(self.pyro_potential, track_gradients=True)
+        elif "slice_np" in method:
+            return partial(self.posterior_potential, track_gradients=False)
+        elif method == "rejection":
+            return partial(self.posterior_potential, track_gradients=True)
         else:
-            return self.np_potential
+            NotImplementedError
 
-    def np_potential(self, theta: np.ndarray) -> ScalarFloat:
-        r"""Return posterior theta log prob. $p(\theta|x)$, $-\infty$ if outside prior."
+    def posterior_potential(
+        self, theta: Union[Tensor, np.array], track_gradients: bool = False
+    ) -> Tensor:
+        """
+        Return posterior theta log prob. $p(\theta|x)$, $-\infty$ if outside prior.
 
         Args:
-            theta: Parameters $\theta$, batch dimension 1.
-
-        Returns:
-            Posterior log probability $\log(p(\theta|x))$.
+            theta:  Parameters $\theta$. If a `transform` is applied, `theta` should be
+                in transformed space.
         """
-        theta = torch.as_tensor(theta, dtype=torch.float32)
-        theta = ensure_theta_batched(theta)
-        num_batch = theta.shape[0]
 
-        # Repeat x over batch dim to match theta batch, accounting for multi-D x.
-        x_repeated = self.x.repeat(num_batch, *(1 for _ in range(self.x.ndim - 1)))
+        # Device is the same for net and prior.
+        transformed_theta = ensure_theta_batched(
+            torch.as_tensor(theta, dtype=torch.float32)
+        ).to(self.device)
+        # Transform `theta` from transformed (i.e. unconstrained) to untransformed
+        # space.
+        theta = self.transform.inv(transformed_theta)
+        log_abs_det = self.transform.log_abs_det_jacobian(theta, transformed_theta)
 
-        with torch.set_grad_enabled(False):
-            target_log_prob = self.posterior_nn.log_prob(
-                inputs=theta.to(self.device),
-                context=x_repeated,
-            )
+        theta_repeated, x_repeated = DirectPosterior._match_theta_and_x_batch_shapes(
+            theta, self.x
+        )
+
+        with torch.set_grad_enabled(track_gradients):
+
+            # Evaluate on device, move back to cpu for comparison with prior.
+            posterior_log_prob = self.posterior_nn.log_prob(theta_repeated, x_repeated)
+            posterior_log_prob_transformed = posterior_log_prob - log_abs_det
+
+            # Force probability to be zero outside prior support.
             in_prior_support = within_support(self.prior, theta)
-            target_log_prob[~in_prior_support] = -float("Inf")
 
-        return target_log_prob
+            posterior_log_prob_transformed = torch.where(
+                in_prior_support,
+                posterior_log_prob_transformed,
+                torch.tensor(float("-inf"), dtype=torch.float32, device=self.device),
+            )
+
+        return posterior_log_prob_transformed
 
     def pyro_potential(
         self, theta: Dict[str, Tensor], track_gradients: bool = False
     ) -> Tensor:
-        r"""Return posterior log prob. of theta $p(\theta|x)$, -inf where outside prior.
+        r"""Return posterior theta log prob. $p(\theta|x)$, $-\infty$ if outside prior."
 
         Args:
-            theta: Parameters $\theta$ (from pyro sampler).
+            theta: Parameters $\theta$ (from pyro sampler). If a `transform` is
+                applied, `theta` should be in transformed space.
 
         Returns:
-            Posterior log probability $p(\theta|x)$, masked outside of prior.
+            Negative posterior log probability $p(\theta|x)$, masked outside of prior.
         """
 
         theta = next(iter(theta.values()))
-
-        with torch.set_grad_enabled(track_gradients):
-            # Notice opposite sign to numpy.
-            # Move theta to device for evaluation.
-            log_prob_posterior = -self.posterior_nn.log_prob(
-                inputs=theta.to(self.device),
-                context=self.x,
-            ).cpu()
-
-        in_prior_support = within_support(self.prior, theta)
-
-        return torch.where(
-            in_prior_support,
-            log_prob_posterior,
-            float("-inf") * torch.ones_like(log_prob_posterior),
-        )
+        return -self.posterior_potential(theta, track_gradients=track_gradients)
