@@ -7,12 +7,13 @@ from warnings import warn
 
 import numpy as np
 import torch
+import torch.distributions.transforms as torch_tf
 from torch import Tensor, nn
 
 from sbi.inference.posteriors.base_posterior import NeuralPosterior
 from sbi.types import Shape
-from sbi.utils import del_entries
-from sbi.utils.torchutils import ScalarFloat, atleast_2d, ensure_theta_batched
+from sbi.utils import del_entries, mcmc_transform, rejection_sample
+from sbi.utils.torchutils import atleast_2d, ensure_theta_batched
 
 
 class RatioBasedPosterior(NeuralPosterior):
@@ -35,8 +36,10 @@ class RatioBasedPosterior(NeuralPosterior):
         neural_net: nn.Module,
         prior,
         x_shape: torch.Size,
+        sample_with: str = "mcmc",
         mcmc_method: str = "slice_np",
         mcmc_parameters: Optional[Dict[str, Any]] = None,
+        rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
         device: str = "cpu",
     ):
         """
@@ -44,7 +47,14 @@ class RatioBasedPosterior(NeuralPosterior):
             method_family: One of snpe, snl, snre_a or snre_b.
             neural_net: A classifier for SNRE, a density estimator for SNPE and SNL.
             prior: Prior distribution with `.log_prob()` and `.sample()`.
-            x_shape: Shape of a single simulator output.
+            x_shape: Shape of the simulated data. It can differ from the
+                observed data the posterior is conditioned on later in the batch
+                dimension. If it differs, the additional entries are interpreted as
+                independent and identically distributed data / trials. I.e., the data is
+                assumed to be generated based on the same (unknown) model parameters or
+                experimental condations.
+            sample_with: Method to use for sampling from the posterior. Must be one of
+                [`mcmc` | `rejection`].
             mcmc_method: Method used for MCMC sampling, one of `slice_np`, `slice`,
                 `hmc`, `nuts`. Currently defaults to `slice_np` for a custom numpy
                 implementation of slice sampling; select `hmc`, `nuts` or `slice` for
@@ -57,6 +67,15 @@ class RatioBasedPosterior(NeuralPosterior):
                 will draw init locations from prior, whereas `sir` will use Sequential-
                 Importance-Resampling using `init_strategy_num_candidates` to find init
                 locations.
+            rejection_sampling_parameters: Dictionary overriding the default parameters
+                for rejection sampling. The following parameters are supported:
+                `proposal` as the proposal distribtution (default is the prior).
+                `max_sampling_batch_size` as the batchsize of samples being drawn from
+                the proposal at every iteration. `num_samples_to_find_max` as the
+                number of samples that are used to find the maximum of the
+                `potential_fn / proposal` ratio. `num_iter_to_find_max` as the number
+                of gradient ascent iterations to find the maximum of that ratio. `m` as
+                multiplier to that ratio.
             device: Training device, e.g., cpu or cuda:0.
         """
         kwargs = del_entries(locals(), entries=("self", "__class__"))
@@ -99,7 +118,7 @@ class RatioBasedPosterior(NeuralPosterior):
             track_gradients=track_gradients,
         )
 
-        return log_ratio.cpu() + self._prior.log_prob(theta)
+        return log_ratio + self._prior.log_prob(theta)
 
     def _warn_log_prob_snre(self) -> None:
         if self._method_family == "snre_a":
@@ -119,9 +138,10 @@ class RatioBasedPosterior(NeuralPosterior):
         sample_shape: Shape = torch.Size(),
         x: Optional[Tensor] = None,
         show_progress_bars: bool = True,
-        sample_with_mcmc: Optional[bool] = None,
+        sample_with: Optional[str] = None,
         mcmc_method: Optional[str] = None,
         mcmc_parameters: Optional[Dict[str, Any]] = None,
+        rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
     ) -> Tensor:
         r"""
         Return samples from posterior distribution $p(\theta|x)$ with MCMC.
@@ -133,40 +153,88 @@ class RatioBasedPosterior(NeuralPosterior):
             x: Conditioning context for posterior $p(\theta|x)$. If not provided,
                 fall back onto `x` passed to `set_default_x()`.
             show_progress_bars: Whether to show sampling progress monitor.
-            sample_with_mcmc: Optional parameter to override `self.sample_with_mcmc`.
+            sample_with: Method to use for sampling from the posterior. Must be one of
+                [`mcmc` | `rejection`].
             mcmc_method: Optional parameter to override `self.mcmc_method`.
             mcmc_parameters: Dictionary overriding the default parameters for MCMC.
-                The following parameters are supported: `thin` to set the thinning
-                factor for the chain, `warmup_steps` to set the initial number of
-                samples to discard, `num_chains` for the number of chains,
+                The following parameters are supported:
+                `thin` to set the thinning factor for the chain.
+                `warmup_steps` to set the initial number of samples to discard.
+                `num_chains` for the number of chains.
                 `init_strategy` for the initialisation strategy for chains; `prior`
                 will draw init locations from prior, whereas `sir` will use Sequential-
                 Importance-Resampling using `init_strategy_num_candidates` to find init
                 locations.
+                `enable_transform` a bool indicating whether MCMC is performed in
+                z-scored (and unconstrained) space.
+            rejection_sampling_parameters: Dictionary overriding the default parameters
+                for rejection sampling. The following parameters are supported:
+                `proposal` as the proposal distribtution (default is the prior).
+                `max_sampling_batch_size` as the batchsize of samples being drawn from
+                the proposal at every iteration.
+                `num_samples_to_find_max` as the number of samples that are used to
+                find the maximum of the `potential_fn / proposal` ratio.
+                `num_iter_to_find_max` as the number of gradient ascent iterations to
+                find the maximum of that ratio.
+                `m` as multiplier to that ratio.
 
         Returns:
             Samples from posterior.
         """
 
-        x, num_samples, mcmc_method, mcmc_parameters = self._prepare_for_sample(
-            x, sample_shape, mcmc_method, mcmc_parameters
-        )
-
         self.net.eval()
 
+        sample_with = sample_with if sample_with is not None else self._sample_with
+        x, num_samples = self._prepare_for_sample(x, sample_shape)
+
         potential_fn_provider = PotentialFunctionProvider()
-        samples = self._sample_posterior_mcmc(
-            num_samples=num_samples,
-            potential_fn=potential_fn_provider(self._prior, self.net, x, mcmc_method),
-            init_fn=self._build_mcmc_init_fn(
-                self._prior,
-                potential_fn_provider(self._prior, self.net, x, "slice_np"),
+        if sample_with == "mcmc":
+            mcmc_method, mcmc_parameters = self._potentially_replace_mcmc_parameters(
+                mcmc_method, mcmc_parameters
+            )
+
+            transform = mcmc_transform(
+                self._prior, device=self._device, **mcmc_parameters
+            )
+
+            transformed_samples = self._sample_posterior_mcmc(
+                num_samples=num_samples,
+                potential_fn=potential_fn_provider(
+                    self._prior, self.net, x, mcmc_method, transform
+                ),
+                init_fn=self._build_mcmc_init_fn(
+                    self._prior,
+                    potential_fn_provider(
+                        self._prior, self.net, x, "slice_np", transform
+                    ),
+                    transform=transform,
+                    **mcmc_parameters,
+                ),
+                mcmc_method=mcmc_method,
+                show_progress_bars=show_progress_bars,
                 **mcmc_parameters,
-            ),
-            mcmc_method=mcmc_method,
-            show_progress_bars=show_progress_bars,
-            **mcmc_parameters,
-        )
+            )
+            samples = transform.inv(transformed_samples)
+        elif sample_with == "rejection":
+            rejection_sampling_parameters = (
+                self._potentially_replace_rejection_parameters(
+                    rejection_sampling_parameters
+                )
+            )
+            if "proposal" not in rejection_sampling_parameters:
+                rejection_sampling_parameters["proposal"] = self._prior
+
+            samples, _ = rejection_sample(
+                potential_fn=potential_fn_provider(
+                    self._prior, self.net, x, "rejection"
+                ),
+                num_samples=num_samples,
+                **rejection_sampling_parameters,
+            )
+        else:
+            raise NameError(
+                "The only implemented sampling methods are `mcmc` and `rejection`."
+            )
 
         self.net.train(True)
 
@@ -178,9 +246,11 @@ class RatioBasedPosterior(NeuralPosterior):
         condition: Tensor,
         dims_to_sample: List[int],
         x: Optional[Tensor] = None,
+        sample_with: str = "mcmc",
         show_progress_bars: bool = True,
         mcmc_method: Optional[str] = None,
         mcmc_parameters: Optional[Dict[str, Any]] = None,
+        rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
     ) -> Tensor:
         r"""
         Return samples from conditional posterior $p(\theta_i|\theta_j, x)$.
@@ -204,16 +274,32 @@ class RatioBasedPosterior(NeuralPosterior):
                 `condition`.
             x: Conditioning context for posterior $p(\theta|x)$. If not provided,
                 fall back onto `x` passed to `set_default_x()`.
+            sample_with: Method to use for sampling from the posterior. Must be one of
+                [`mcmc` | `rejection`]. In this method, the value of
+                `self._sample_with` will be ignored.
             show_progress_bars: Whether to show sampling progress monitor.
             mcmc_method: Optional parameter to override `self.mcmc_method`.
             mcmc_parameters: Dictionary overriding the default parameters for MCMC.
-                The following parameters are supported: `thin` to set the thinning
-                factor for the chain, `warmup_steps` to set the initial number of
-                samples to discard, `num_chains` for the number of chains,
+                The following parameters are supported:
+                `thin` to set the thinning factor for the chain.
+                `warmup_steps` to set the initial number of samples to discard.
+                `num_chains` for the number of chains.
                 `init_strategy` for the initialisation strategy for chains; `prior`
                 will draw init locations from prior, whereas `sir` will use Sequential-
                 Importance-Resampling using `init_strategy_num_candidates` to find init
                 locations.
+                `enable_transform` a bool indicating whether MCMC is performed in
+                z-scored (and unconstrained) space.
+            rejection_sampling_parameters: Dictionary overriding the default parameters
+                for rejection sampling. The following parameters are supported:
+                `proposal` as the proposal distribtution (default is the prior).
+                `max_sampling_batch_size` as the batchsize of samples being drawn from
+                the proposal at every iteration.
+                `num_samples_to_find_max` as the number of samples that are used to
+                find the maximum of the `potential_fn / proposal` ratio.
+                `num_iter_to_find_max` as the number of gradient ascent iterations to
+                find the maximum of that ratio.
+                `m` as multiplier to that ratio.
 
         Returns:
             Samples from conditional posterior.
@@ -225,9 +311,11 @@ class RatioBasedPosterior(NeuralPosterior):
             condition,
             dims_to_sample,
             x,
+            sample_with,
             show_progress_bars,
             mcmc_method,
             mcmc_parameters,
+            rejection_sampling_parameters,
         )
 
     def map(
@@ -260,10 +348,6 @@ class RatioBasedPosterior(NeuralPosterior):
                 fall back onto `x` passed to `set_default_x()`.
             num_iter: Maximum Number of optimization steps that the algorithm takes
                 to find the MAP.
-            early_stop_at: If `None`, it will optimize for `max_num_iter` iterations.
-                If `float`, the optimization will stop as soon as the steps taken by
-                the optimizer are smaller than `early_stop_at` times the standard
-                deviation of the initial guesses.
             learning_rate: Learning rate of the optimizer.
             init_method: How to select the starting parameters for the optimization. If
                 it is a string, it can be either [`posterior`, `prior`], which samples
@@ -324,10 +408,7 @@ class RatioBasedPosterior(NeuralPosterior):
 
     @staticmethod
     def _log_ratios_over_trials(
-        x: Tensor,
-        theta: Tensor,
-        net: nn.Module,
-        track_gradients: bool = False,
+        x: Tensor, theta: Tensor, net: nn.Module, track_gradients: bool = False
     ) -> Tensor:
         r"""Return log ratios summed over iid trials of `x`.
 
@@ -383,7 +464,7 @@ class PotentialFunctionProvider:
      most current trained posterior neural net.
 
     Returns:
-        Potential function for use by either numpy or pyro sampler
+        Potential function for use by either numpy, pyro, or rejection sampler.
     """
 
     def __call__(
@@ -391,18 +472,19 @@ class PotentialFunctionProvider:
         prior,
         classifier: nn.Module,
         x: Tensor,
-        mcmc_method: str,
+        method: str,
+        transform: torch_tf.Transform = torch_tf.identity_transform,
     ) -> Callable:
         r"""Return potential function for posterior $p(\theta|x)$.
 
-        Switch on numpy or pyro potential function based on `mcmc_method`.
+        Switch on numpy or pyro potential function based on `method`.
 
         Args:
             prior: Prior distribution that can be evaluated.
             classifier: Binary classifier approximating the likelihood up to a constant.
 
             x: Conditioning variable for posterior $p(\theta|x)$.
-            mcmc_method: One of `slice_np`, `slice`, `hmc` or `nuts`.
+            method: One of `slice_np`, `slice`, `hmc` or `nuts`, `rejection`.
 
         Returns:
             Potential function for sampler.
@@ -412,34 +494,52 @@ class PotentialFunctionProvider:
         self.prior = prior
         self.device = next(classifier.parameters()).device
         self.x = atleast_2d(x).to(self.device)
+        self.transform = transform
 
-        if mcmc_method == "slice":
+        if method == "slice":
             return partial(self.pyro_potential, track_gradients=False)
-        elif mcmc_method in ("hmc", "nuts"):
+        elif method in ("hmc", "nuts"):
             return partial(self.pyro_potential, track_gradients=True)
+        elif "slice_np" in method:
+            return partial(self.posterior_potential, track_gradients=False)
+        elif method == "rejection":
+            return partial(self.posterior_potential, track_gradients=True)
         else:
-            return self.np_potential
+            NotImplementedError
 
-    def np_potential(self, theta: np.array) -> ScalarFloat:
-        """Return potential for Numpy slice sampler."
+    def posterior_potential(
+        self, theta: Union[Tensor, np.array], track_gradients: bool = False
+    ) -> Tensor:
+        """Returns the unnormalized posterior log-probability.
 
-        For numpy MCMC samplers this is the unnormalized posterior log prob.
+        This is the potential used in the numpy slice sampler and in rejection sampling.
 
         Args:
-            theta: Parameters $\theta$, batch dimension 1.
+            theta: Parameters $\theta$, batch dimension 1. If a `transform` is applied,
+                `theta` should be in transformed space.
 
         Returns:
             Posterior log probability of theta.
         """
-        theta = torch.as_tensor(theta, dtype=torch.float32)
-        theta = ensure_theta_batched(theta)
+
+        # Device is the same for net and prior.
+        transformed_theta = ensure_theta_batched(
+            torch.as_tensor(theta, dtype=torch.float32)
+        ).to(self.device)
+        # Transform `theta` from transformed (i.e. unconstrained) to untransformed
+        # space.
+        theta = self.transform.inv(transformed_theta)
+        log_abs_det = self.transform.log_abs_det_jacobian(theta, transformed_theta)
 
         log_ratio = RatioBasedPosterior._log_ratios_over_trials(
-            self.x, theta.to(self.device), self.classifier, track_gradients=False
+            self.x,
+            theta,
+            self.classifier,
+            track_gradients=track_gradients,
         )
-
-        # Notice opposite sign to pyro potential.
-        return log_ratio.cpu() + self.prior.log_prob(theta)
+        posterior_potential = log_ratio + self.prior.log_prob(theta)
+        posterior_potential_transformed = posterior_potential - log_abs_det
+        return posterior_potential_transformed
 
     def pyro_potential(
         self, theta: Dict[str, Tensor], track_gradients: bool = False
@@ -450,8 +550,9 @@ class PotentialFunctionProvider:
 
         Args:
             theta: Parameters $\theta$. The tensor's shape will be
-             (1, shape_of_single_theta) if running a single chain or just
-             (shape_of_single_theta) for multiple chains.
+                (1, shape_of_single_theta) if running a single chain or just
+                (shape_of_single_theta) for multiple chains. If a `transform` is
+                applied, `theta` should be in transformed space.
 
         Returns:
             Potential $-(\log r(x_o, \theta) + \log p(\theta))$.
@@ -459,14 +560,5 @@ class PotentialFunctionProvider:
 
         theta = next(iter(theta.values()))
 
-        # Theta and x should have shape (1, dim).
-        theta = ensure_theta_batched(theta)
-
-        log_ratio = RatioBasedPosterior._log_ratios_over_trials(
-            self.x,
-            theta.to(self.device),
-            self.classifier,
-            track_gradients=track_gradients,
-        )
-
-        return -(log_ratio.cpu() + self.prior.log_prob(theta))
+        # Note the minus to match the pyro potential function requirements.
+        return -self.posterior_potential(theta, track_gradients=track_gradients)
