@@ -95,12 +95,7 @@ class SNPE_A(PosteriorEstimator):
         # to pass arguments between functions, and that's implicit state management.
         kwargs = utils.del_entries(
             locals(),
-            entries=(
-                "self",
-                "__class__",
-                "unused_args",
-                "num_components",
-            ),
+            entries=("self", "__class__", "unused_args", "num_components"),
         )
         super().__init__(**kwargs)
 
@@ -119,7 +114,7 @@ class SNPE_A(PosteriorEstimator):
         retrain_from_scratch_each_round: bool = False,
         show_train_summary: bool = False,
         dataloader_kwargs: Optional[Dict] = None,
-        component_perturbation: float = 1e-5,
+        component_perturbation: float = 5e-3,
     ) -> DirectPosterior:
         r"""
         Return density estimator that approximates the proposal posterior $\tilde{p}(\theta|x)$.
@@ -162,7 +157,8 @@ class SNPE_A(PosteriorEstimator):
                 and validation dataloaders (like, e.g., a collate_fn)
             component_perturbation: The standard deviation applied to all weights and
                 biases when, in the last round, the Mixture of Gaussians is build from
-                a single Gaussian.
+                a single Gaussian. This value can be problem-specific and also depends
+                on the number of mixture components.
         Returns:
             Density estimator that approximates the distribution $p(\theta|x)$.
         """
@@ -223,7 +219,6 @@ class SNPE_A(PosteriorEstimator):
         self,
         density_estimator: Optional[TorchModule] = None,
         rejection_sampling_parameters: Optional[Dict[str, Any]] = None,
-        allow_precision_correction: bool = False,
     ) -> DirectPosterior:
         r"""
         Build posterior from the neural density estimator.
@@ -251,10 +246,6 @@ class SNPE_A(PosteriorEstimator):
                 `max_sampling_batch_size` to set the batch size for drawing new
                 samples from the candidate distribution, e.g., the posterior. Larger
                 batch size speeds up sampling.
-            allow_precision_correction: Whether to add a diagonal with the smallest
-                eigenvalue in every entry in case the precision matrix becomes
-                ill-conditioned.
-
         Returns:
             Posterior $p(\theta|x)$  with `.sample()` and `.log_prob()` methods.
         """
@@ -287,10 +278,7 @@ class SNPE_A(PosteriorEstimator):
 
         # Create the SNPE_A_MDN
         wrapped_density_estimator = SNPE_A_MDN(
-            flow=density_estimator,
-            proposal=proposal,
-            prior=self._prior,
-            allow_precision_correction=allow_precision_correction,
+            flow=density_estimator, proposal=proposal, prior=self._prior
         )
 
         self._posterior = DirectPosterior(
@@ -298,8 +286,8 @@ class SNPE_A(PosteriorEstimator):
             neural_net=wrapped_density_estimator,
             prior=self._prior,
             x_shape=self._x_shape,
+            sample_with="rejection",
             rejection_sampling_parameters=rejection_sampling_parameters,
-            sample_with_mcmc=False,
             device=device,
         )
 
@@ -385,7 +373,6 @@ class SNPE_A_MDN(nn.Module):
         flow: flows.Flow,
         proposal: Union["utils.BoxUniform", MultivariateNormal, "SNPE_A_MDN"],
         prior: Any,
-        allow_precision_correction: bool = False,
     ):
         """Constructor.
 
@@ -393,15 +380,12 @@ class SNPE_A_MDN(nn.Module):
             flow: The trained normalizing flow, passed when building the posterior.
             proposal: The proposal distribution.
             prior: The prior distribution.
-            allow_precision_correction: Add a diagonal with the smallest eigenvalue in
-                every entry in case the precision matrix becomes ill-conditioned.
         """
         # Call nn.Module's constructor.
         super().__init__()
 
         self._neural_net = flow
         self._prior = prior
-        self._allow_precision_correction = allow_precision_correction
 
         # Set the proposal using the `default_x`.
         if isinstance(proposal, (utils.BoxUniform, MultivariateNormal)):
@@ -436,10 +420,7 @@ class SNPE_A_MDN(nn.Module):
 
             # Compute the log_prob of theta under the product.
             log_prob_proposal_posterior = utils.sbiutils.mog_log_prob(
-                theta,
-                logits_pp,
-                m_pp,
-                prec_pp,
+                theta, logits_pp, m_pp, prec_pp
             )
             utils.assert_all_finite(
                 log_prob_proposal_posterior, "proposal posterior eval"
@@ -538,12 +519,7 @@ class SNPE_A_MDN(nn.Module):
 
         # Compute the MoG parameters of the posterior.
         logits_p, m_p, prec_p, cov_p = self._proposal_posterior_transformation(
-            logits_pp,
-            m_pp,
-            prec_pp,
-            norm_logits_d,
-            m_d,
-            prec_d,
+            logits_pp, m_pp, prec_pp, norm_logits_d, m_d, prec_d
         )
         return logits_p, m_p, prec_p
 
@@ -586,11 +562,7 @@ class SNPE_A_MDN(nn.Module):
         )
 
         means_post = self._means_posterior(
-            covariances_post,
-            means_pp,
-            precisions_pp,
-            means_d,
-            precisions_d,
+            covariances_post, means_pp, precisions_pp, means_d, precisions_d
         )
 
         logits_post = SNPE_A_MDN._logits_posterior(
@@ -656,9 +628,8 @@ class SNPE_A_MDN(nn.Module):
         prior will not be exactly have mean=0 and std=1.
         """
         if self.z_score_theta:
-            # Default to cpu to avoid mismatch with prior (always on cpu) below.
-            scale = self._neural_net._transform._transforms[0]._scale.cpu()
-            shift = self._neural_net._transform._transforms[0]._shift.cpu()
+            scale = self._neural_net._transform._transforms[0]._scale
+            shift = self._neural_net._transform._transforms[0]._shift
 
             # Following the definition of the linear transform in
             # `standardizing_transform` in `sbiutils.py`:
@@ -720,18 +691,18 @@ class SNPE_A_MDN(nn.Module):
         for batches in precisions_pp:
             for pprior in batches:
                 eig_pprior = torch.symeig(pprior, eigenvectors=False).eigenvalues
-                assert (
-                    eig_pprior > 0
-                ).all(), (
-                    "The precision matrix of the proposal is not positive definite!"
-                )
+                if not (eig_pprior > 0).all():
+                    raise AssertionError(
+                        "The precision matrix of the proposal is not positive definite!"
+                    )
         for batches in precisions_d:
             for d in batches:
                 eig_d = torch.symeig(d, eigenvectors=False).eigenvalues
-                assert (
-                    eig_d > 0
-                ).all(), """The precision matrix of the density estimator is not
-                positive definite!"""
+                if not (eig_d > 0).all():
+                    raise AssertionError(
+                        "The precision matrix of the density estimator is not "
+                        "positive definite!"
+                    )
 
         precisions_pp_rep = precisions_pp.repeat_interleave(num_comps_d, dim=1)
         precisions_d_rep = precisions_d.repeat(1, num_comps_p, 1, 1)
@@ -745,26 +716,14 @@ class SNPE_A_MDN(nn.Module):
             for idx_comp, pp in enumerate(batches):
                 eig_pp = torch.symeig(pp, eigenvectors=False).eigenvalues
                 if not (eig_pp > 0).all():
-                    if self._allow_precision_correction:
-                        # Shift the eigenvalues to be at minimum 1e-6.
-                        precisions_p[idx_batch, idx_comp] = pp - torch.eye(
-                            pp.shape[0]
-                        ) * (min(eig_pp) - 1e-6)
-                        warnings.warn(
-                            "The precision matrix of a posterior has not been positive "
-                            "definite at least once. Added diagonal entries with the "
-                            "smallest eigenvalue to 1e-6."
-                        )
-
-                    else:
-                        # Fail when encountering an ill-conditioned precision matrix.
-                        raise AssertionError(
-                            "The precision matrix of a posterior is not positive "
-                            "definite! This is a known issue for SNPE-A. Either try a "
-                            "different parameter setting or pass"
-                            "`allow_precision_correction=True` when calling "
-                            "`build_posterior()`."
-                        )
+                    raise AssertionError(
+                        "The precision matrix of a posterior is not positive "
+                        "definite! This is a known issue for SNPE-A. Either try a "
+                        "different parameter setting, e.g. a different number of "
+                        "mixture components (when contracting SNPE-A), or a different "
+                        "value for the parameter perturbation (when building the "
+                        "posterior)."
+                    )
 
         covariances_p = torch.inverse(precisions_p)
         return precisions_p, covariances_p
